// Adapted from: Jos Stam, "Real-Time Fluid Dynamics for Games". Proceedings of the Game Developer Conference, March 2003.
// For details on how the individual functions work, please consult the above publication.

#include "FluidSimComponent.h"

#include "Engine/VolumeTexture.h"

UFluidSimComponent::UFluidSimComponent()
{
	PrimaryComponentTick.bCanEverTick = true;
}

// Called when the game starts
void UFluidSimComponent::BeginPlay()
{
	Super::BeginPlay();

	CreateVolumeTransient();

	// Expand the actual computation grid to account for edge cases
	DensityGridSize = Size + 2;

	// Create data arrays, fill them with zeroes
	Density.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	DensityPrevious.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);

	VelocityX.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	VelocityXPrevious.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	
	VelocityY.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	VelocityYPrevious.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	
	VelocityZ.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	VelocityZPrevious.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
}

// Called every frame
void UFluidSimComponent::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

	// Run Stam's stable fluid solver
	if (bSimulating)
	{
		VelocityStep(DeltaTime);
		DensityStep(DeltaTime);
	}

	UE_LOG(LogClass, Warning, TEXT("Value at center: %f"), Density[GetIndex(Size / 2 + 1, Size / 2 + 1, Size / 2 + 1)]);

	// Lock texture data so we can update it
	uint8* ByteArray = (uint8*)VolumeTexture->PlatformData->Mips[0].BulkData.Lock(LOCK_READ_WRITE);

	// Loop through all our cells and convert their values to color.
	// Only takes into account values between 0.0-1.0 and maps them to 0-255
	int Current = 0;
	for (int i = 1; i <= Size; i++)
	{
		for (int j = 1; j <= Size; j++)
		{
			for (int k = 1; k <= Size; k++)
			{
				int val = 255 * Density[GetIndex(k, j, i)] / 1.0f;
				
				val = std::min(255, std::max(0, val));

				// 4 values: R G B A, a byte each
				ByteArray[Current + 0] = val;
				ByteArray[Current + 1] = val;
				ByteArray[Current + 2] = val;
				ByteArray[Current + 3] = val;

				Current += 4;
			}
		}
	}

	// We're done updating - unlock the data and tell UE to update the texture on the GPU side.
	VolumeTexture->PlatformData->Mips[0].BulkData.Unlock();
	VolumeTexture->UpdateResource();

	// Clear data for next tick.
	FMemory::Memset(DensityPrevious.GetData(), 0, DensityPrevious.GetAllocatedSize());
	FMemory::Memset(VelocityXPrevious.GetData(), 0, VelocityXPrevious.GetAllocatedSize());
	FMemory::Memset(VelocityYPrevious.GetData(), 0, VelocityYPrevious.GetAllocatedSize());
	FMemory::Memset(VelocityZPrevious.GetData(), 0, VelocityZPrevious.GetAllocatedSize());
}

void UFluidSimComponent::InjectFluid(FVector Position, float Amount)
{
	// Map the unit vector to a specific cell.
	int x = Position.X * Size;
	int y = Position.Y * Size;
	int z = Position.Z * Size;

	x = std::min(Size, std::max(1, x));
	y = std::min(Size, std::max(1, y));
	z = std::min(Size, std::max(1, z));

	DensityPrevious[GetIndex(x, y, z)] += Amount;
}

void UFluidSimComponent::ApplyForce(FVector Position, FVector Force)
{
	int x = Position.X * Size;
	int y = Position.Y * Size;
	int z = Position.Z * Size;

	x = std::min(Size, std::max(1, x));
	y = std::min(Size, std::max(1, y));
	z = std::min(Size, std::max(1, z));

	VelocityXPrevious[GetIndex(x, y, z)] = Force.X;
	VelocityYPrevious[GetIndex(x, y, z)] = Force.Y;
	VelocityZPrevious[GetIndex(x, y, z)] = Force.Z;
}

void UFluidSimComponent::CreateVolumeTransient()
{
	// First create the object itself.
	VolumeTexture = NewObject<UVolumeTexture>(GetTransientPackage(), NAME_None, RF_Transient);
	
	if (!VolumeTexture->PlatformData)
	{
		VolumeTexture->PlatformData = new FTexturePlatformData();
	}
	// Set Dimensions and Pixel format.
	VolumeTexture->PlatformData->SizeX = Size;
	VolumeTexture->PlatformData->SizeY = Size;
	VolumeTexture->PlatformData->SetNumSlices(Size);
	VolumeTexture->PlatformData->PixelFormat = EPixelFormat::PF_R8G8B8A8;
	// Set sRGB and streaming to false.
	VolumeTexture->SRGB = false;
	VolumeTexture->NeverStream = true;
	
	int PixelByteSize = 8 * 4;
	const long long TotalSize = (long long) Size * Size * Size * PixelByteSize;

	// Create the one and only mip in this texture.
	FTexture2DMipMap* Mip = new FTexture2DMipMap();
	Mip->SizeX = Size;
	Mip->SizeY = Size;
	Mip->SizeZ = Size;

	Mip->BulkData.Lock(LOCK_READ_WRITE);
	// Allocate memory in the mip and copy the actual texture data inside
	uint8* ByteArray = (uint8*) Mip->BulkData.Realloc(TotalSize);

	// Memset to zero
	FMemory::Memset(ByteArray, 0, TotalSize);

	Mip->BulkData.Unlock();

	// Newly created Volume textures have this null'd
	if (!VolumeTexture->PlatformData)
	{
		VolumeTexture->PlatformData = new FTexturePlatformData();
	}
	// Add the new MIP to the list of mips.
	VolumeTexture->PlatformData->Mips.Add(Mip);

	// Tell Unreal to update the rest.
	VolumeTexture->UpdateResource();
}

void UFluidSimComponent::AddSources(TArray<float>& Cells, const TArray<float>& Sources, float DeltaTime)
{
	int GridSize = DensityGridSize * DensityGridSize * DensityGridSize;
	for (int i = 0; i < GridSize; i++)
	{
		Cells[i] += Sources[i] * DeltaTime; 
	}
}

void UFluidSimComponent::Diffuse(int BoundaryType, TArray<float>& X, 
	TArray<float>& X0, float Diffusion, float DeltaTime)
{
	// The actual factor is the input one scaled by the time step and scaled to the size of our simulation grid.
	float Diff = DeltaTime * Diffusion * Size * Size * Size;

	// TODO: Use the Gauss-Seidel relaxation method to calculate diffusion.
	// Don't forget to call SetBounds on X to prevent spill in each solver step!
}

void UFluidSimComponent::DiffuseBad(int BoundaryType, TArray<float>& X, 
	TArray<float>& X0, float Diffusion, float DeltaTime)
{
	float Diff = DeltaTime * Diffusion * Size * Size * Size;
 
	for (int i = 1; i <= Size; i++)
	{ 
		for (int j = 1; j <= Size; j++)
		{
			for (int k = 1; k <= Size; k++)
			{
				// Naive forward approach. Faster than the stable solution and works for sufficiently small
				// time steps (and diffusion factors), but can easily become unstable (imagine a pause elsewhere in the
				// game, resulting in a very high time step).
				X[GetIndex(i, j, k)] = X0[GetIndex(i, j, k)]
				+ Diff*(X0[GetIndex(i + 1, j, k)]
				+ X0[GetIndex(i - 1, j, k)]
				+ X0[GetIndex(i, j + 1, k)]
				+ X0[GetIndex(i, j - 1, k)]
				+ X0[GetIndex(i, j, k + 1)]
				+ X0[GetIndex(i, j, k - 1)]
				- 6 * X0[GetIndex(i, j, k)]);
			}
		}
	}
	SetBounds(BoundaryType, X);
}

void UFluidSimComponent::Advect(int BoundaryType, TArray<float>& D, TArray<float>& D0, TArray<float>& VX, TArray<float>& VY, TArray<float>& VZ, float DeltaTime)
{
	float Delta = DeltaTime * Size;
	
	for (int i = 1; i <= Size; i++)
	{
		for (int j = 1; j <= Size; j++)
		{
			for (int k = 1; k <= Size; k++)
			{
				// Find source position
				float x = i - Delta * VX[GetIndex(i, j, k)];
				float y = j - Delta * VY[GetIndex(i, j, k)];
				float z = k - Delta * VZ[GetIndex(i, j, k)];

				// Make sure we're sufficiently inside the grid
				if (x < 0.5f) x = 0.5f; if (x > Size + 0.5f) x = Size + 0.5f;
				if (y < 0.5f) y = 0.5f; if (y > Size + 0.5f) y = Size + 0.5f;
				if (z < 0.5f) z = 0.5f; if (z > Size + 0.5f) z = Size + 0.5f;

				// Compute neighbouring cells and their contributions
				int i0 = static_cast<int>(x);
				int i1 = i0 + 1;
				
				int j0 = static_cast<int>(y);
				int j1 = j0 + 1;
				
				int k0 = static_cast<int>(z);
				int k1 = k0 + 1;

				float s1 = x - i0;
				float s0 = 1 - s1;
				
				float t1 = y - j0;
				float t0 = 1 - t1;
				
				float u1 = z - k0;
				float u0 = 1 - u1;

				// Linear interpolation between the 8 neighbours
				D[GetIndex(i, j, k)] =
					s0 * (t0 * (u0 * D0[GetIndex(i0, j0, k0)] + u1 * D0[GetIndex(i0, j0, k1)]) +
						  t1 * (u0 * D0[GetIndex(i0, j1, k0)] + u1 * D0[GetIndex(i0, j1, k1)])) +
					s1 * (t0 * (u0 * D0[GetIndex(i1, j0, k0)] + u1 * D0[GetIndex(i1, j0, k1)]) +
						  t1 * (u0 * D0[GetIndex(i1, j1, k0)] + u1 * D0[GetIndex(i1, j1, k1)]));
			}
		}
	}
	SetBounds(BoundaryType, D);
}

void UFluidSimComponent::DensityStep(float DeltaTime)
{
	// Stam's original code uses pointer swapping - here we just swap the variable names instead.
	// Swapping is used instead of additional temporary arrays. Swapping means result is always computed to a "new" array.
	AddSources(Density, DensityPrevious, DeltaTime);
	// Try both variants with different parameters, see how the bad variant breaks quickly with higher values.
	if (bUseLinearSolver)
	{
		Diffuse(0, DensityPrevious, Density, DiffusionFactor, DeltaTime);
	}
	else
	{
		DiffuseBad(0, DensityPrevious, Density, DiffusionFactor, DeltaTime);
	}
	Advect(0, Density, DensityPrevious, VelocityX, VelocityY, VelocityZ, DeltaTime);
}

void UFluidSimComponent::Project(TArray<float>& Vx, TArray<float>& Vy, TArray<float>& Vz, TArray<float>& Temp,
	TArray<float>& Divisions)
{
	float InvSize = 1.0f/Size; 
	for (int i = 1; i <= Size; i++)
	{ 
		for (int j = 1; j <= Size; j++)
		{
			for (int k = 1; k <= Size; k++)
			{
				Divisions[GetIndex(i, j, k)] = -1.0f / 3.0f * InvSize * (
					Vx[GetIndex(i + 1, j, k)] - Vx[GetIndex(i - 1, j, k)] +
					Vy[GetIndex(i, j + 1, k)] - Vy[GetIndex(i, j - 1, k)] +
					Vz[GetIndex(i, j, k + 1) - Vz[GetIndex(i, j, k - 1)]]);
				Temp[GetIndex(i, j, k)] = 0;
			}
		} 
	}
	SetBounds(0, Divisions); SetBounds(0, Temp);

	for (int Step = 0; Step < LinearSolverSteps; Step++)
	{ 
		for (int i = 1; i <= Size; i++)
		{
			for (int j = 1; j <= Size; j++)
			{
				for (int k = 1; k <= Size; k++)
				{
					Temp[GetIndex(i, j, k)] = (Divisions[GetIndex(i, j, k)] +
						Temp[GetIndex(i - 1, j, k)] +
						Temp[GetIndex(i + 1, j, k)] +
						Temp[GetIndex(i, j + 1, k)] +
						Temp[GetIndex(i, j - 1, k)] +
						Temp[GetIndex(i, j, k + 1)] +
						Temp[GetIndex(i, j, k - 1)]) / 6;	
				} 
			}
		}
		SetBounds(0, Temp);
	} 
 
	for (int i = 1; i <= Size; i++)
	{
		for (int j = 1;j <= Size; j++)
		{
			for (int k = 1; k <= Size; k++)
			{
				Vx[GetIndex(i, j, k)] -= 0.5f*(Temp[GetIndex(i + 1, j, k)] - Temp[GetIndex(i - 1, j, k)])/InvSize;
				Vy[GetIndex(i, j, k)] -= 0.5f*(Temp[GetIndex(i, j + 1, k)] - Temp[GetIndex(i, j - 1, k)])/InvSize;
				Vz[GetIndex(i, j, k)] -= 0.5f*(Temp[GetIndex(i, j, k + 1)] - Temp[GetIndex(i, j, k - 1)])/InvSize;
			}
		} 
	}
	SetBounds(1, Vx); SetBounds(2, Vy); SetBounds(3, Vz);
}

void UFluidSimComponent::VelocityStep(float DeltaTime)
{
	// Stam's original code uses pointer swapping - here we just swap the variable names instead.
	// Each vector component is processed separately as per Stam's original algorithm. We could also
	// use Unreal's vectors for this, or use this separation for parallel computation.
	AddSources(VelocityX, VelocityXPrevious, DeltaTime);
	AddSources(VelocityY, VelocityYPrevious, DeltaTime);
	AddSources(VelocityZ, VelocityZPrevious, DeltaTime);

	Diffuse(1, VelocityXPrevious, VelocityX, Viscosity, DeltaTime);
	Diffuse(2, VelocityYPrevious, VelocityY, Viscosity, DeltaTime);
	Diffuse(3, VelocityZPrevious, VelocityZ, Viscosity, DeltaTime);

	Project(VelocityXPrevious, VelocityYPrevious, VelocityZPrevious, VelocityX, VelocityY);
	
	Advect(1, VelocityX, VelocityXPrevious, VelocityXPrevious, VelocityYPrevious, VelocityZPrevious, DeltaTime);
	Advect(2, VelocityY, VelocityYPrevious, VelocityXPrevious, VelocityYPrevious, VelocityZPrevious, DeltaTime);
	Advect(3, VelocityZ, VelocityZPrevious, VelocityXPrevious, VelocityYPrevious, VelocityZPrevious, DeltaTime);

	Project(VelocityX, VelocityY, VelocityZ, VelocityXPrevious, VelocityYPrevious);
}

void UFluidSimComponent::SetBounds(int Type, TArray<float>& Data)
{
	const int L = Size + 1;
	
	for (int i = 1; i <= Size; i++)
	{
		// First take care of the walls of the fluid's bounding box
		// The type parameter value is used here to determine when to "reflect" the vector along the wall
		for (int j = 1; j <= Size; j++)
		{
			Data[GetIndex(0, i, j)] = Type == 1 ? -Data[GetIndex(1, i, j)] : Data[GetIndex(1, i, j)];
			Data[GetIndex(L, i, j)] = Type == 1 ? -Data[GetIndex(L - 1, i, j)] : Data[GetIndex(L - 1, i, j)];
			
			Data[GetIndex(i, 0, j)] = Type == 2 ? -Data[GetIndex(i, 1, j)] : Data[GetIndex(i, 1, j)];
			Data[GetIndex(i, L, j)] = Type == 2 ? -Data[GetIndex(i, L - 1, j)] : Data[GetIndex(i, L - 1, j)];
			
			Data[GetIndex(i, j, 0)] = Type == 3 ? -Data[GetIndex(i, j, 1)] : Data[GetIndex(i, j, 1)];
			Data[GetIndex(i, j, L)] = Type == 3 ? -Data[GetIndex(i, j, L - 1)] : Data[GetIndex(i, j, L - 1)];
		}

		// Next set limiting values along the 6 edges of the box
		Data[GetIndex(0, i, L)] = 0.5f * (Data[GetIndex(1    , i, L)] + Data[GetIndex(0, i, L - 1)]);
		Data[GetIndex(L, i, L)] = 0.5f * (Data[GetIndex(L - 1, i, L)] + Data[GetIndex(L, i, L - 1)]);
		Data[GetIndex(0, i, 0)] = 0.5f * (Data[GetIndex(1    , i, 0)] + Data[GetIndex(0, i, 1    )]);
		Data[GetIndex(L, i, 0)] = 0.5f * (Data[GetIndex(L - 1, i, 0)] + Data[GetIndex(L, i, 1    )]);
		
		Data[GetIndex(i, 0, L)] = 0.5f * (Data[GetIndex(i, 1    , L)] + Data[GetIndex(i, 0, L - 1)]);
		Data[GetIndex(i, L, L)] = 0.5f * (Data[GetIndex(i, L - 1, L)] + Data[GetIndex(i, L, L - 1)]);
		Data[GetIndex(i, 0, 0)] = 0.5f * (Data[GetIndex(i, 1    , 0)] + Data[GetIndex(i, 0, 1    )]);
		Data[GetIndex(i, L, 0)] = 0.5f * (Data[GetIndex(i, L - 1, 0)] + Data[GetIndex(i, L, 1    )]);
		
		Data[GetIndex(0, L, i)] = 0.5f * (Data[GetIndex(1    , L, i)] + Data[GetIndex(0, L - 1, i)]);
		Data[GetIndex(L, L, i)] = 0.5f * (Data[GetIndex(L - 1, L, i)] + Data[GetIndex(L, L - 1, i)]);
		Data[GetIndex(0, 0, i)] = 0.5f * (Data[GetIndex(1    , 0, i)] + Data[GetIndex(0, 0    , i)]);
		Data[GetIndex(L, 0, i)] = 0.5f * (Data[GetIndex(L - 1, 0, i)] + Data[GetIndex(L, 0    , i)]);
	}

	// Finally handle the 8 corners of the box
	constexpr float Third = 1.0 / 3.0;
	Data[GetIndex(0, 0, 0)] = Third * (Data[GetIndex(1    , 0, 0)] + Data[GetIndex(0, 1    , 0)] + Data[GetIndex(0, 0, 1    )]);
	Data[GetIndex(0, L, 0)] = Third * (Data[GetIndex(1    , L, 0)] + Data[GetIndex(0, L - 1, 0)] + Data[GetIndex(0, L, 1    )]);
	Data[GetIndex(L, 0, 0)] = Third * (Data[GetIndex(L - 1, 0, 0)] + Data[GetIndex(L, 1    , 0)] + Data[GetIndex(L, 0, 1    )]);
	Data[GetIndex(L, L, 0)] = Third * (Data[GetIndex(L - 1, L, 0)] + Data[GetIndex(L, L - 1, 0)] + Data[GetIndex(L, L, 1    )]);
	Data[GetIndex(0, 0, L)] = Third * (Data[GetIndex(1    , 0, L)] + Data[GetIndex(0, 1    , L)] + Data[GetIndex(0, 0, L - 1)]);
	Data[GetIndex(0, L, L)] = Third * (Data[GetIndex(1    , L, L)] + Data[GetIndex(0, L - 1, L)] + Data[GetIndex(0, L, L - 1)]);
	Data[GetIndex(L, 0, L)] = Third * (Data[GetIndex(L - 1, 0, L)] + Data[GetIndex(L, 1    , L)] + Data[GetIndex(L, 0, L - 1)]);
	Data[GetIndex(L, L, L)] = Third * (Data[GetIndex(L - 1, L, L)] + Data[GetIndex(L, L - 1, L)] + Data[GetIndex(L, L, L - 1)]);
}

