#pragma once

#include "CoreMinimal.h"
#include "Components/SceneComponent.h"
#include "FluidSimComponent.generated.h"


UCLASS( ClassGroup=(Custom), meta=(BlueprintSpawnableComponent) )
class CV4_TEKUTINY_API UFluidSimComponent : public USceneComponent
{
	GENERATED_BODY()

public:	
	// Sets default values for this component's properties
	UFluidSimComponent();
	
	UPROPERTY(EditAnywhere, Category="Volume Parameters")
	int Size = 32;

	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category="Volume Parameters")
	UVolumeTexture* VolumeTexture;

	UPROPERTY(EditAnywhere, Category="Fluid Simulation")
	int LinearSolverSteps = 20;
	UPROPERTY(EditAnywhere, Category="Fluid Simulation")
	float DiffusionFactor = 0.0f;
	UPROPERTY(EditAnywhere, Category="Fluid Simulation")
	float Viscosity = 0.0f;
	UPROPERTY(EditAnywhere, Category="Fluid Simulation")
	bool bSimulating = true;
	UPROPERTY(EditAnywhere, Category="Fluid Simulation")
	bool bUseLinearSolver = true;

	UFUNCTION(BlueprintCallable, Category="Fluid Simulation")
	// Injects new fluid into the simulation at the given position (expects a unit vector)
	void InjectFluid(FVector Position, float Amount);

	UFUNCTION(BlueprintCallable, Category="Fluid Simulation")
	// Adds a force into the simulation's force field (expects a unit vector)
	void ApplyForce(FVector Position, FVector Force);
	
protected:
	virtual void BeginPlay() override;

public:	
	virtual void TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction) override;

private:
	// Creates a runtime VolumeTexture through Unreal API. This texture exists only when playing the game. It
	// has a single mip level with the full resolution of our simulation cell grid (minus the edge cells). 
	void CreateVolumeTransient();
	
	// Converts a set of coordinates to a cell array index. Simplifies working with a single-dimensional array
	// representing a multi-dimensional structure.
	uint32 GetIndex(int x, int y, int z) const { return DensityGridSize * DensityGridSize * z + DensityGridSize * y + x; }
	
	// Add time-scaled values from the source array s to the destination array x.
	void AddSources(TArray<float>& Cells, const TArray<float>& Sources, float DeltaTime);
	
	// High-level part of the simulation which uses the above functions to compute velocity field changes in a
	// single step of the simulation.
	void VelocityStep(float DeltaTime);
	
	// High-level part of the simulation which uses the above functions to compute density changes in a single
	// step of the simulation.
	void DensityStep(float DeltaTime);

	
	// Diffuses the values in the source array. Density/velocity each cell has is shared with its neighbours to
	// simulate natural diffusion of gasses (and the force fields which affect them, i.e. wind). Used several times in
	// the simulation. This version computes what density each cell should have if diffusion happened in reverse. This
	// way each cell is a separate variable, resulting in a set of linear equations which can be solved using the
	// iterative Gauss-Seidel method.
	// btype Type of border limiting values to apply (used for velocities) - passed when calling SetBounds
	void Diffuse(int BoundaryType, TArray<float>& X, TArray<float>& X0, float Diffusion, float DeltaTime);
	
	// Same as Diffuse, but calculates diffusion forward in time. The algorithm is more straightforward but is
	// very unstable, especially for larger diffusion factor and time step.
	// btype Type of border limiting values to apply (used for velocities) - passed when calling SetBounds
	void DiffuseBad(int BoundaryType, TArray<float>& X, TArray<float>& X0, float Diffusion, float DeltaTime);
	
	// Computes the flow of the densities/forces due to advection from the vector field. This makes the fluid
	// move across the cells. Computation is done similarly to the Diffusion step - backwards in time. We look at a position
	// and linearly work out which cell the value came from. Since the resulting value is linearly interpolated from the
	// computed position's neighbouring cells, this step also does pseudo-diffusion.
	// btype Type of border limiting values to apply (used for velocities) - passed when calling SetBounds 
	void Advect(int BoundaryType, TArray<float>& D, TArray<float>& D0, TArray<float>& VX, TArray<float>& VY, TArray<float>& VZ, float DeltaTime);

	
	// This step ensures conservation of mass in our fluid's velocity field. This is done by using Hodge decomposition.
	// First, a gradient is computed for the velocity field (into p), which is subtracted from our velocity field.
	// The result is a mass conserving field which is saved to the input velocity field arrays.
	void Project(TArray<float>& Vx, TArray<float>& Vy, TArray<float>& Vz, TArray<float>& Temp, TArray<float>& Divisions);
	
	
	// Applies edge conditions to the additional grid layer which is used in this simulation. This implementation
	// encloses the fluid in a solid box.
	// type Type of edge binding conditions to apply 
	void SetBounds(int Type, TArray<float>& Data);

	uint32 DensityGridSize;
	TArray<float> Density;
	TArray<float> DensityPrevious;

	TArray<float> VelocityX;
	TArray<float> VelocityY;
	TArray<float> VelocityZ;
	
	TArray<float> VelocityXPrevious;
	TArray<float> VelocityYPrevious;
	TArray<float> VelocityZPrevious;
};
